from __future__ import annotations

from typing import AsyncContextManager, AsyncIterator

from nats.aio.client import Client as NATSClient
from nats.aio.msg import Msg

from .executor import RequestManyExecutor
from .iterator import RequestManyIterator


class Client(NATSClient):
    def __init__(
        self,
        max_wait: float = 0.5,
    ) -> None:
        super().__init__()
        self.max_wait = max_wait

    def request_many_iter(
        self,
        subject: str,
        payload: bytes | None = None,
        headers: dict[str, str] | None = None,
        reply_inbox: str | None = None,
        max_wait: float | None = None,
        max_count: int | None = None,
        max_interval: float | None = None,
        stop_on_sentinel: bool = False,
    ) -> AsyncContextManager[AsyncIterator[Msg]]:
        """Request many responses from the same subject.

        The iterator exits without raising an error when no responses are received.

        Responses are received until one of the following conditions is met:

        - max_wait seconds have passed.
        - max_count responses have been received.
        - max_interval seconds have passed between responses.
        - A sentinel message is received and stop_on_sentinel is True.

        When any of the condition is met, the async iterator yielded by the
        context manager raises StopAsyncIteration on the next iteration.

        The subscription is started when entering the async context manager and
        stopped when exiting.

        Args:
            subject: The subject to send the request to.
            payload: The payload to send with the request.
            headers: The headers to send with the request.
            reply_inbox: The inbox to receive the responses in. A new inbox is created if None.
            max_wait: The maximum amount of time to wait for responses. 1 second by default.
            max_count: The maximum number of responses to accept. No limit by default.
            max_interval: The maximum amount of time between responses. No limit by default.
            stop_on_sentinel: Whether to stop when a sentinel message is received. False by default.
        """
        inbox = reply_inbox or self.new_inbox()
        return RequestManyIterator(
            self,
            subject,
            payload=payload,
            headers=headers,
            inbox=inbox,
            max_wait=max_wait,
            max_count=max_count,
            max_interval=max_interval,
            stop_on_sentinel=stop_on_sentinel,
        )

    async def request_many(
        self,
        subject: str,
        payload: bytes | None = None,
        headers: dict[str, str] | None = None,
        reply_inbox: str | None = None,
        max_wait: float | None = None,
        max_count: int | None = None,
        max_interval: float | None = None,
        stop_on_sentinel: bool = False,
    ) -> list[Msg]:
        """Request many responses from the same subject.

        This function does not raise an error when no responses are received.

        Responses are received until one of the following conditions is met:

        - max_wait seconds have passed.
        - max_count responses have been received.
        - max_interval seconds have passed between responses.
        - A sentinel message is received and stop_on_sentinel is True.

        Subscription is always stopped when the function returns.

        Args:
            subject: The subject to send the request to.
            payload: The payload to send with the request.
            headers: The headers to send with the request.
            reply_inbox: The inbox to receive the responses in. A new inbox is created if None.
            max_wait: The maximum amount of time to wait for responses. 1 second by default.
            max_count: The maximum number of responses to accept. No limit by default.
            max_interval: The maximum amount of time between responses. No limit by default.
            stop_on_sentinel: Whether to stop when a sentinel message is received. False by default.
        """
        executor = RequestManyExecutor(self, max_wait)
        return await executor(
            subject,
            reply_inbox=reply_inbox,
            payload=payload,
            headers=headers,
            max_wait=max_wait,
            max_count=max_count,
            max_interval=max_interval,
            stop_on_sentinel=stop_on_sentinel,
        )



from __future__ import annotations

import asyncio
from typing import Any, AsyncIterator

from nats.aio.client import Client
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription
from nats.errors import BadSubscriptionError


class RequestManyIterator:

    def __init__(
        self,
        nc: Client,
        subject: str,
        inbox: str,
        payload: bytes | None = None,
        headers: dict[str, str] | None = None,
        max_wait: float | None = None,
        max_interval: float | None = None,
        max_count: int | None = None,
        stop_on_sentinel: bool = False,
    ) -> None:
        """Request many responses from the same subject.

        Request is sent when entering the async context manager and unsubscribed when exiting.

        The async iterator yieled by the context manager do not raise an
        error when no responses are received.

        Responses are received until one of the following conditions is met:

        - max_wait seconds have passed.
        - max_count responses have been received.
        - max_interval seconds have passed between responses.
        - A sentinel message is received and stop_on_sentinel is True.

        When any of the condition is met, the async iterator raises StopAsyncIteration on
        the next call to __anext__, and the subscription is unsubscribed on exit.

        Args:
            subject: The subject to send the request to.
            payload: The payload to send with the request.
            headers: The headers to send with the request.
            inbox: The inbox to receive the responses in. A new inbox is created if None.
            max_wait: The maximum amount of time to wait for responses. Default max wait can be configured at the instance level.
            max_count: The maximum number of responses to accept. No limit by default.
            max_interval: The maximum amount of time between responses. No limit by default.
            stop_on_sentinel: Whether to stop when a sentinel message is received. False by default.
        """
        if max_wait is None and max_interval is None:
            max_wait = 0.5
        # Save all the arguments as instance variables.
        self.nc = nc
        self.subject = subject
        self.payload = payload
        self.headers = headers
        self.inbox = inbox
        self.max_wait = max_wait
        self.max_count = max_count
        self.max_interval = max_interval
        self.stop_on_sentinel = stop_on_sentinel
        # Initialize the state of the request many iterator
        self._sub: Subscription | None = None
        self._iterator: AsyncIterator[Msg] | None = None
        self._did_unsubscribe = False
        self._total_received = 0
        self._last_received = asyncio.get_event_loop().time()
        self._tasks: list[asyncio.Task[object]] = []
        self._pending_task: asyncio.Task[Msg] | None = None

    def __aiter__(self) -> RequestManyIterator:
        """RequestManyIterator is an asynchronous iterator."""
        return self

    async def __anext__(self) -> Msg:
        """Return the next message or raise StopAsyncIteration."""
        if not self._sub:
            raise RuntimeError(
                "RequestManyIterator must be used as an async context manager"
            )
        # Exit early if we've already unsubscribed
        if self._did_unsubscribe:
            raise StopAsyncIteration
        # Exit early if we received all the messages
        if self.max_count and self._total_received == self.max_count:
            if self._sub and not self._did_unsubscribe:
                self._did_unsubscribe = True
                await _unsubscribe(self._sub)
            raise StopAsyncIteration
        # Create a task to wait for the next message
        task: asyncio.Task[Msg] = asyncio.create_task(self._iterator.__anext__())  # type: ignore
        self._pending_task = task
        # Wait for the next message or any of the other tasks to complete
        await asyncio.wait(
            [self._pending_task, *self._tasks],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if self._pending_task.cancelled():
            raise StopAsyncIteration
        if not self._pending_task.done():
            self._pending_task.cancel()
            raise StopAsyncIteration
        # if err := self._pending_task.exception():
        #     raise err
        # This will raise an exception if an error occurred within the task
        msg = self._pending_task.result()
        # Check message headers
        # If the message is a 503 error, raise StopAsyncIteration
        if msg.headers and msg.headers.get("Status") == "503":
            raise StopAsyncIteration
        # Always increment the total received count
        self._total_received += 1
        # Check if this is a sentinel message
        if self.stop_on_sentinel and msg.data == b"":
            if self._sub and not self._did_unsubscribe:
                self._did_unsubscribe = True
                await _unsubscribe(self._sub)
            # In which case, raise StopAsyncIteration and don't return the message
            raise StopAsyncIteration
        # Return the message
        return msg

    async def __aenter__(self) -> RequestManyIterator:
        """Start the subscription and publish the request."""
        # Start the subscription
        sub = await self.nc.subscribe(  # pyright: ignore[reportUnknownMemberType]
            self.inbox,
            max_msgs=self.max_count or 0,
        )
        # Save the subscription and the iterator
        self._iterator = sub.messages
        self._sub = sub
        # Add a task to wait for the max_wait time if needed
        if self.max_wait:
            self._tasks.append(asyncio.create_task(asyncio.sleep(self.max_wait)))
        # Add a task to check the interval if needed
        if self.max_interval:
            interval = self.max_interval

            async def check_interval() -> None:
                while True:
                    await asyncio.sleep(interval)
                    if asyncio.get_event_loop().time() - self._last_received > interval:
                        if self._sub and not self._did_unsubscribe:
                            self._did_unsubscribe = True
                            await _unsubscribe(self._sub)
                        return

            self._tasks.append(asyncio.create_task(check_interval()))

        # Publish the request
        await self.nc.publish(
            self.subject, self.payload or b"", reply=self.inbox, headers=self.headers
        )
        # At this point the subscription is ready and all tasks are submitted
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        """Unsubscribe from the inbox and cancel all the tasks."""
        for task in self._tasks:
            if not task.done():
                task.cancel()
        if self._pending_task and not self._pending_task.done():
            self._pending_task.cancel()
        if self._sub and not self._did_unsubscribe:
            await _unsubscribe(self._sub)


async def _unsubscribe(sub: Subscription) -> None:
    try:
        await sub.unsubscribe()
    except BadSubscriptionError:
        # It's possible that auto-unsubscribe has already been called.
        pass



from __future__ import annotations

import asyncio
import contextlib
from typing import AsyncIterator, Awaitable, Callable

import pytest
import pytest_asyncio
from async_timeout import timeout
from nats.aio.msg import Msg
from nats_contrib.test_server import NATSD

from nats_contrib.request_many import Client as NATS


@pytest.mark.asyncio
class RequestManyIteratorTestSetup:
    @pytest_asyncio.fixture(autouse=True)
    async def setup(self, nats_server: NATSD, nats_client: NATS) -> AsyncIterator[None]:
        self.nats_server = nats_server
        self.nats_client = nats_client
        self.test_stack = contextlib.AsyncExitStack()
        async with self.test_stack:
            yield None


class TestRequestManyIterator(RequestManyIteratorTestSetup):
    """Test suite for request_many_iter method."""

    @pytest.mark.parametrize(
        ["max_wait", "max_interval"],
        [
            (None, None),
            (0.1, None),
            (None, 0.1),
            (0.1, 0.1),
        ],
    )
    async def test_request_many_iter_no_responder(
        self, max_wait: float | None, max_interval: float | None
    ) -> None:
        async with timeout(0.05):
            # Request many without a responder and big timeout
            async with self.nats_client.request_many_iter(
                "not.a.subject.with.listener",
                max_wait=max_wait,
                max_interval=max_interval,
            ) as messages:
                assert len([msg async for msg in messages]) == 0

    async def test_request_many_iter_single_responder(self) -> None:
        # Define a handler function
        async def responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        # Create service
        sub = await self.nats_client.subscribe("subject.with.listener", cb=responder)
        # Unsubscribe when done
        self.test_stack.push_async_callback(sub.unsubscribe)
        async with timeout(0.2):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1
            ) as replies:
                msg, *rest = [msg async for msg in replies]

        # One message should be received
        assert len(rest) == 0
        # Message should be the expected one
        assert msg.data == b"OK"

    async def test_request_many_iter_multiple_responders(self) -> None:
        # Define a handler function
        def make_responder(index: int) -> Callable[[Msg], Awaitable[None]]:
            async def responder(msg: Msg) -> None:
                await msg.respond(f"{index}".encode())

            return responder

        # Create services
        subs = [
            await self.nats_client.subscribe(
                "subject.with.listener", cb=make_responder(i)
            )
            for i in range(5)
        ]
        # Unsubscribe when done
        for sub in subs:
            self.test_stack.push_async_callback(sub.unsubscribe)
        # Request many with multiple responders
        # Task should not reach timeout
        async with timeout(0.2):
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1
            ) as replies:
                msgs = [msg async for msg in replies]
        # All messages should be received
        assert len(msgs) == 5
        # All messages should be the expected one
        assert set(msg.data for msg in msgs) == {b"0", b"1", b"2", b"3", b"4"}

    async def test_request_many_iter_max_count_not_reached(self) -> None:
        # Define a handler function
        async def responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        # Create service
        sub = await self.nats_client.subscribe("subject.with.listener", cb=responder)
        # Unsubscribe when done
        self.test_stack.push_async_callback(sub.unsubscribe)
        async with timeout(0.2):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1, max_count=5
            ) as replies:
                msgs = [msg async for msg in replies]
        # One message should be received
        assert len(msgs) == 1
        # Message should be the expected one
        assert msgs[0].data == b"OK"

    async def test_request_many_iter_max_count_reached(self) -> None:
        # Define a handler function
        async def responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        # Create service
        sub = await self.nats_client.subscribe("subject.with.listener", cb=responder)
        # Unsubscribe when done
        self.test_stack.push_async_callback(sub.unsubscribe)
        async with timeout(0.2):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1, max_count=1
            ) as replies:
                msgs = [msg async for msg in replies]
        # One message should be received
        assert len(msgs) == 1
        # Message should be the expected one
        assert msgs[0].data == b"OK"

    async def test_request_many_iter_max_count_reached_multiple_responders(
        self,
    ) -> None:
        # Define a handler function
        def make_responder(index: int) -> Callable[[Msg], Awaitable[None]]:
            async def responder(msg: Msg) -> None:
                await msg.respond(f"{index}".encode())

            return responder

        # Create services
        subs = [
            await self.nats_client.subscribe(
                "subject.with.listener", cb=make_responder(i)
            )
            for i in range(3)
        ]
        # Unsubscribe when done
        for sub in subs:
            self.test_stack.push_async_callback(sub.unsubscribe)
        # Request many with multiple responders
        # Task should not reach timeout
        async with timeout(0.2):
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1, max_count=3
            ) as replies:
                msgs = [msg async for msg in replies]
        # All messages should be received
        assert len(msgs) == 3
        # All messages should be the expected one
        assert set(msg.data for msg in msgs) == {b"0", b"1", b"2"}

    async def test_request_many_max_interval_reached(self) -> None:
        # Define a handler function
        async def first_responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        async def second_responder(msg: Msg) -> None:
            await asyncio.sleep(0.2)
            await msg.respond(b"OK")

        # Create service
        for resp in [first_responder, second_responder]:
            sub = await self.nats_client.subscribe("subject.with.listener", cb=resp)
            # Unsubscribe when done
            self.test_stack.push_async_callback(sub.unsubscribe)

        async with timeout(0.1):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_interval=0.01
            ) as replies:
                msgs = [msg async for msg in replies]
        # One message should be received
        assert len(msgs) == 1
        # Message should be the expected one
        assert msgs[0].data == b"OK"

    async def test_request_many_custom_inbox(self) -> None:
        # Define a handler function
        async def responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        # Create service
        sub = await self.nats_client.subscribe("subject.with.listener", cb=responder)
        # Unsubscribe when done
        self.test_stack.push_async_callback(sub.unsubscribe)
        async with timeout(0.2):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1, reply_inbox="custom.inbox"
            ) as replies:
                msgs = [msg async for msg in replies]
        # One message should be received
        assert len(msgs) == 1
        # Message should be the expected one
        assert msgs[0].data == b"OK"
        assert msgs[0].subject == "custom.inbox"

    async def test_request_many_iter_sentinel(self) -> None:
        async def first_responder(msg: Msg) -> None:
            await msg.respond(b"OK")

        async def other_responder(msg: Msg) -> None:
            await asyncio.sleep(1e-3)
            await msg.respond(b"")

        # Create service
        sub = await self.nats_client.subscribe(
            "subject.with.listener", cb=first_responder
        )
        # Unsubscribe when done
        self.test_stack.push_async_callback(sub.unsubscribe)
        for _ in range(5):
            sub = await self.nats_client.subscribe(
                "subject.with.listener", cb=other_responder
            )
            # Unsubscribe when done
            self.test_stack.push_async_callback(sub.unsubscribe)

        async with timeout(0.2):
            # Request many with a single responder
            async with self.nats_client.request_many_iter(
                "subject.with.listener", max_wait=0.1, stop_on_sentinel=True
            ) as replies:
                msgs = [msg async for msg in replies]
        # One message should be received
        assert len(msgs) == 1
        # Message should be the expected one
        assert msgs[0].data == b"OK"



from __future__ import annotations

import asyncio

from nats.aio.client import Client
from nats.aio.msg import Msg
from nats.errors import BadSubscriptionError


class RequestManyExecutor:
    def __init__(
        self,
        nc: Client,
        max_wait: float | None = None,
    ) -> None:
        self.nc = nc
        self.max_wait = max_wait or 0.5

    async def __call__(
        self,
        subject: str,
        reply_inbox: str | None = None,
        payload: bytes | None = None,
        headers: dict[str, str] | None = None,
        max_wait: float | None = None,
        max_count: int | None = None,
        max_interval: float | None = None,
        stop_on_sentinel: bool = False,
    ) -> list[Msg]:
        """Request many responses from the same subject.

        This function does not raise an error when no responses are received.

        Responses are received until one of the following conditions is met:

        - max_wait seconds have passed.
        - max_count responses have been received.
        - max_interval seconds have passed between responses.
        - A sentinel message is received and stop_on_sentinel is True.

        Args:
            subject: The subject to send the request to.
            payload: The payload to send with the request.
            headers: The headers to send with the request.
            reply_inbox: The inbox to receive the responses in. A new inbox is created if None.
            max_wait: The maximum amount of time to wait for responses. Default max wait can be configured at the instance level.
            max_count: The maximum number of responses to accept. No limit by default.
            max_interval: The maximum amount of time between responses. No limit by default.
            stop_on_sentinel: Whether to stop when a sentinel message is received. False by default.
        """
        if max_wait is None and max_interval is None:
            max_wait = self.max_wait
        # Create an inbox for the responses if one wasn't provided.
        if reply_inbox is None:
            reply_inbox = self.nc.new_inbox()
        # Create an empty list to store the responses.
        responses: list[Msg] = []
        # Get the event loop
        loop = asyncio.get_event_loop()
        # Create an event to signal when the request is complete.
        event = asyncio.Event()
        # Create a marker to indicate that a message was received
        # and the interval has passed.
        last_received = loop.time()

        # Define a callback to handle the responses.
        async def callback(msg: Msg) -> None:
            # Update the last received time.
            nonlocal last_received
            last_received = loop.time()
            # Check message headers
            # If the message is a 503 error, set the event and return.
            if msg.headers and msg.headers.get("Status") == "503":
                event.set()
                return
            # If we're stopping on a sentinel message, check for it
            # and don't append the message to the list of responses.
            if stop_on_sentinel and msg.data == b"":
                event.set()
                return
            # In all other cases, append the message to the list of responses.
            responses.append(msg)
            # And check if we've received all the responses.
            if len(responses) == max_count:
                event.set()

        # Subscribe to the inbox.
        sub = await self.nc.subscribe(  # pyright: ignore[reportUnknownMemberType]
            reply_inbox,
            cb=callback,
            max_msgs=max_count or 0,
        )
        # Initialize a list of tasks to wait for.
        tasks: list[asyncio.Task[object]] = []
        # Enter try/finally clause to ensure that the subscription is
        # unsubscribed from even if an error occurs.
        try:
            # Create task to wait for the stop event.
            tasks.append(asyncio.create_task(event.wait()))

            # Add a task to wait for the max_wait time if needed
            if max_wait:
                tasks.append(asyncio.create_task(asyncio.sleep(max_wait)))

            # Add a task to check the interval if needed
            if max_interval:

                async def check_interval() -> None:
                    nonlocal last_received
                    while True:
                        await asyncio.sleep(max_interval)
                        if loop.time() - last_received > max_interval:
                            event.set()
                            return

                tasks.append(asyncio.create_task(check_interval()))

            # At this point the subscription is ready and all tasks are submitted
            # Publish the request.
            await self.nc.publish(
                subject, payload or b"", reply=reply_inbox, headers=headers
            )
            # Wait for the first task to complete.
            await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED,
            )
        # Always cancel tasks and unsubscribe from the inbox.
        finally:
            # Cancel the remaining tasks as soon as first one completes.
            for task in tasks:
                if not task.done():
                    task.cancel()
            # Unsubscribe from the inbox.
            try:
                await sub.unsubscribe()
            except BadSubscriptionError:
                # It's possible that auto-unsubscribe has already been called.
                pass

        # Return the list of responses.
        return responses



from __future__ import annotations

from typing import Any, AsyncContextManager, AsyncIterator, Callable, Generic, TypeVar

T = TypeVar("T")
R = TypeVar("R")


def transform(
    source: AsyncContextManager[AsyncIterator[T]],
    map: Callable[[T], R],
) -> AsyncContextManager[AsyncIterator[R]]:
    """Create a new async context manager which will
    yield an async iterator that applies the map function to each value
    yielded by the source async iterator.

    It is useful for example to transform the return value of the
    `request_many_iter` method.
    """
    return TransformAsyncIterator(source, map)


class TransformAsyncIterator(Generic[T, R]):
    def __init__(
        self,
        source: AsyncContextManager[AsyncIterator[T]],
        map: Callable[[T], R],
    ) -> None:
        self.factory = source
        self.iterator: AsyncIterator[T] | None = None
        self.transform = map

    def __aiter__(self) -> TransformAsyncIterator[T, R]:
        return self

    async def __anext__(self) -> R:
        if not self.iterator:
            raise RuntimeError(
                "TransformAsyncIterator must be used as an async context manager"
            )
        next_value = await self.iterator.__anext__()
        return self.transform(next_value)

    async def __aenter__(self) -> AsyncIterator[R]:
        self.iterator = await self.factory.__aenter__()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.factory.__aexit__(*args, **kwargs)



from .client import Client
from .executor import RequestManyExecutor
from .iterator import RequestManyIterator
from .utils import transform

__all__ = ["Client", "RequestManyExecutor", "RequestManyIterator", "transform"]



from __future__ import annotations

import pytest

from nats_contrib.request_many.utils import transform


class StubIteratorContext:
    def __init__(self, length: int) -> None:
        self.length = length
        self.count = 0

    async def __aenter__(self) -> StubIteratorContext:
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        pass

    def __aiter__(self) -> StubIteratorContext:
        return self

    async def __anext__(self) -> int:
        if self.count < self.length:
            self.count += 1
            return self.count
        raise StopAsyncIteration


@pytest.mark.asyncio
class TestTransform:
    """Test suite for transfor function."""

    async def test_transform(self):

        ctx = transform(StubIteratorContext(3), lambda x: x * 2)
        async with ctx as iterator:
            assert [x async for x in iterator] == [2, 4, 6]

    @pytest.mark.asyncio
    async def test_transform_bad_usage(self):
        with pytest.raises(RuntimeError) as exc:
            ctx = transform(StubIteratorContext(3), lambda x: x * 2)
            async for msg in ctx:  # type: ignore
                pass

        assert (
            str(exc.value)
            == "TransformAsyncIterator must be used as an async context manager"
        )
